/*
 * Copyright (C) 2010-2013 jeanfi@gmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>

#include <json/json.h>

#include "http.h"
#include "log.h"
#include "ttrss_ws.h"
#include "url.h"

static char *session_id;
static char *session_url;
static char *session_user;
static char *session_pwd;

static pthread_mutex_t lock;
static struct http_session *session;

void ws_request_add_att_str(json_object *rq, const char *k, const char *str)
{
	json_object_object_add(rq, k, json_object_new_string(str));
}

void ws_request_add_att_int(json_object *rq, const char *k, int v)
{
	json_object_object_add(rq, k, json_object_new_int(v));
}

void ws_request_add_att_bool(json_object *rq, const char *k, int v)
{
	json_object_object_add(rq, k, json_object_new_boolean(v));
}

struct json_object *ws_request_new(const char *op)
{
	struct json_object *rq;

	rq = json_object_new_object();

	ws_request_add_att_str(rq, "op", op);

	if (session_id)
		ws_request_add_att_str(rq, "sid", session_id);

	return rq;
}

void ws_set_config(const char *url, const char *user, const char *pwd)
{
	char *tmp;

	if (session_url && !strcmp(session_url, url)
	    && session_user && !strcmp(session_user, user)
	    && session_pwd && !strcmp(session_pwd, pwd))
		return ;

	free(session_id);
	session_id = NULL;

	free(session_user);
	session_user = strdup(user);

	free(session_pwd);
	session_pwd = strdup(pwd);

	tmp = url_normalize(url);
	free(session_url);
	session_url = malloc(strlen(tmp) + strlen("/api/") + 1);
	strcpy(session_url, tmp);
	strcat(session_url, "/api/");
	free(tmp);
}

struct json_object *ws_reply_get_content(struct json_object *rp)
{
	log_debug("ws_reply_get_content");
	return json_object_object_get(rp, "content");
}

static const char *ws_reply_get_error(struct json_object *content)
{
	struct json_object *jerror;

	log_debug("ws_reply_get_error");

	if (json_object_get_type(content) != json_type_object)
		return NULL;

	jerror = json_object_object_get(content, "error");

	if (!jerror)
		return NULL;

	return json_object_get_string(jerror);
}

static struct json_object *
execute(struct http_session *sess, struct json_object *rq, char **err)
{
	struct json_object *rp, *content;
	const char *str;

	rp = http_json_get(sess, session_url, rq);

	content = NULL;

	if (rp) {
		content = ws_reply_get_content(rp);

		if (content) {
			str = ws_reply_get_error(content);

			if (str) {
				*err = strdup(str);
				content = NULL;
			} else {
				json_object_get(content);
			}
		}

		json_object_put(rp);
	}

	return content;
}

struct json_object *ws_execute(struct json_object *rq)
{
	char *err;
	struct json_object *result;

	log_debug("ws_execute()");
	pthread_mutex_lock(&lock);
	log_debug("ws_execute() lock");

	err = NULL;
	result = execute(session, rq, &err);

	if (err) {
		log_debug("ws_execute(): error=%s", err);

		if (!strcmp(err, "NOT_LOGGED_IN")) {
			ws_open_session();
			result = execute(session, rq, NULL);
		}

		free(err);
	}

	log_debug("ws_execute() unlock");
	pthread_mutex_unlock(&lock);

	log_debug("ws_execute()");

	return result;
}

int ws_get_api_version()
{
	struct json_object *rp, *rq;
	int v;

	rq = ws_request_new("getApiLevel");

	rp = ws_execute(rq);

	json_object_put(rq);

	if (rp) {
		v = json_object_get_int(json_object_object_get(rp, "level"));

		json_object_put(rp);
	} else {
		v = 0;
	}

	return v;
}

char *ws_login()
{
	struct json_object *rq, *rp, *j;
	char *str;

	rq = ws_request_new("login");
	ws_request_add_att_str(rq, "user", session_user);
	ws_request_add_att_str(rq, "password", session_pwd);

	rp = ws_execute(rq);
	json_object_put(rq);

	str = NULL;
	if (rp) {
		j = json_object_object_get(rp, "session_id");

		if (j)
			str = strdup(json_object_get_string(j));

		json_object_put(rp);
	}

	return str;
}

int ws_open_session()
{
	int version, result;

	log_debug("ws_open_session()");

	if (session_id)
		free(session_id);

	session_id = NULL;
	session_id = ws_login();

	if (session_id) {
		version = ws_get_api_version();
		log_debug("API version= %d", version);

		if (version > 0) {
			result = 1;
		} else {
			free(session_id);
			session_id = NULL;
			result = 0;
		}
	} else {
		result =  0;
	}

	return result;
}

char *ws_get_article_content(int id)
{
	struct json_object *rp, *rq, *content, *item;
	char *str;

	rq = ws_request_new("getArticle");
	ws_request_add_att_int(rq, "article_id", id);

	rp = ws_execute(rq);

	json_object_put(rq);

	str = NULL;

	if (rp) {
		item = json_object_array_get_idx(rp, 0);

		if (item) {
			content = json_object_object_get(item, "content");
			str = strdup(json_object_get_string(content));
		}

		json_object_put(rp);
	}

	return str;
}

int ws_update_headlines(struct feed *feed)
{
	struct json_object *rp, *rq, *jheadline, *j;
	int i, n, hid;
	struct headline *h, **tmp;
	const char *title, *url;

	rq = ws_request_new("getHeadlines");
	ws_request_add_att_int(rq, "feed_id", feed->id);
	ws_request_add_att_bool(rq, "show_excerpt", 1);

	rp = ws_execute(rq);

	json_object_put(rq);

	if (rp) {
		n = json_object_array_length(rp);
		for (i = 0; i < n; i++) {
			jheadline = json_object_array_get_idx(rp, i);

			j = json_object_object_get(jheadline, "id");
			hid = json_object_get_int(j);
			h = feed_get_headline(feed, hid);

			if (!h) {
				j = json_object_object_get(jheadline, "title");
				title = json_object_get_string(j);

				j = json_object_object_get(jheadline, "link");
				url = json_object_get_string(j);

				h = headline_new(hid, url, title);

				j = json_object_object_get(jheadline,
							   "excerpt");
				h->excerpt = strdup(json_object_get_string(j));

				j = json_object_object_get(jheadline,
							   "updated");
				h->date = json_object_get_int(j);

				tmp = headlines_add(feed->headlines, h);
				if (feed->headlines)
					free(feed->headlines);
				feed->headlines = tmp;
			}

			j = json_object_object_get(jheadline, "unread");
			h->unread = json_object_get_boolean(j);
		}

		if (!feed->headlines) {
			feed->headlines = malloc(sizeof(struct headline *));
			*(feed->headlines) = NULL;
		}

		json_object_put(rp);
		return 1;
	} else {
		return 0;
	}
}

static struct feed **
feeds_update(struct feed **feeds, struct json_object *jarray)
{
	int i, n, id;
	struct json_object *jfeed, *j;
	const char *url, *title;
	struct feed *feed, **tmp;

	if (jarray) {
		n = json_object_array_length(jarray);

		for (i = 0; i < n; i++) {
			jfeed = json_object_array_get_idx(jarray, i);

			j = json_object_object_get(jfeed, "id");
			id = json_object_get_int(j);

			feed = feeds_get_feed(feeds, id);

			if (!feed) {
				j = json_object_object_get(jfeed, "title");
				title = json_object_get_string(j);

				j = json_object_object_get(jfeed, "feed_url");
				url = json_object_get_string(j);

				feed = feed_new(id, url, title);

				tmp = feeds_add(feeds, feed);
				free(feeds);
				feeds = tmp;
			}

			j = json_object_object_get(jfeed, "unread");
			feed->unread = json_object_get_int(j);

			ws_update_headlines(feed);
		}
	}

	return feeds;
}

struct feed **ws_update_feeds(struct feed **feeds)
{
	struct json_object *rp, *rq;

	log_debug("ws_update_feeds()");

	rq = ws_request_new("getFeeds");
	ws_request_add_att_int(rq, "cat_id", 0);

	rp = ws_execute(rq);

	if (rp) {
		feeds = feeds_update(feeds, rp);
		json_object_put(rp);
	}

	ws_request_add_att_int(rq, "cat_id", -3);
	rp = ws_execute(rq);
	json_object_put(rq);

	if (rp) {
		feeds = feeds_update(feeds, rp);
		json_object_put(rp);
	}

	log_debug("ws_update_feeds() done");

	return feeds;
}

struct json_object *ws_request_new_set_article_unread(int id, int unread)
{
	struct json_object *rq;

	rq = ws_request_new("updateArticle");
	json_object_object_add(rq, "article_ids", json_object_new_int(id));
	json_object_object_add(rq, "field", json_object_new_int(2));
	json_object_object_add(rq, "mode", json_object_new_int(unread));

	return rq;
}

void ws_set_article_unread(int id, int unread)
{
	struct json_object *rp, *rq;

	log_debug("ws_set_article_unread(%d,%d)", id, unread);

	rq = ws_request_new_set_article_unread(id, unread);

	rp = ws_execute(rq);

	json_object_put(rq);

	if (rp)
		json_object_put(rp);

	log_debug("ws_set_article_unread(%d,%d) done", id, unread);
}

void ws_init()
{
	pthread_mutexattr_t attr;
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&lock, &attr);

	session = http_session_new();
}
